@@ -6,9 +6,9 @@ package ChemOnomatopist::Chain;
 use strict;
 use warnings;
 
-use ChemOnomatopist;
 use ChemOnomatopist::Chain::Amide;
 use ChemOnomatopist::Chain::Amine;
+use ChemOnomatopist::Chain::Bicycle;
 use ChemOnomatopist::Chain::Ether;
 use ChemOnomatopist::Charge;
 use ChemOnomatopist::Elements qw( %elements );
@@ -17,13 +17,21 @@ use ChemOnomatopist::Group::Ether;
 use ChemOnomatopist::Isotope;
 use ChemOnomatopist::Name::Part::AlkaneANSuffix;
 use ChemOnomatopist::Name::Part::Isotope;
+use ChemOnomatopist::Name::Part::Stereodescriptor;
+use ChemOnomatopist::Util qw( atomic_number );
 use ChemOnomatopist::Util::SMILES qw( path_SMILES );
+use Chemistry::OpenSMILES qw(
+    is_chiral_tetrahedral
+    is_cis_trans_bond
+    is_double_bond
+);
+use Chemistry::OpenSMILES::Writer;
 use Graph::Traversal::DFS;
 use List::Util qw( all any first none sum0 uniq );
 use Scalar::Util qw( blessed );
 use Set::Object qw( set );
 
-sub vertices();
+sub vertices(@);
 
 sub new
 {
@@ -96,10 +104,17 @@ sub substituents()
     return @substituents;
 }
 
-sub vertices()
+sub vertices(@)
 {
-    my( $self ) = @_;
-    return @{$self->{vertices}};
+    my( $self, @vertices_new ) = @_;
+    my @vertices = @{$self->{vertices}};
+    if( @vertices_new ) {
+        $self->{vertices} = \@vertices_new;
+        # TODO: Invalidate all cache
+        delete $self->{bonds};
+        delete $self->{heteroatom_positions};
+    }
+    return @vertices;
 }
 
 # Properties
@@ -202,8 +217,8 @@ sub nonstandard_valence_positions()
     my @nonstandard_valence_positions;
     for (0..$#vertices) {
         next if blessed $vertices[$_];
-        next if ChemOnomatopist::element( $vertices[$_] ) eq 'C';
-        next if ChemOnomatopist::element( $vertices[$_] ) eq 'N';
+        next if ChemOnomatopist::Util::element( $vertices[$_] ) eq 'C';
+        next if ChemOnomatopist::Util::element( $vertices[$_] ) eq 'N';
         next unless exists $vertices[$_]->{valence};
         push @nonstandard_valence_positions, $_;
     }
@@ -212,6 +227,22 @@ sub nonstandard_valence_positions()
     return @nonstandard_valence_positions;
 }
 
+sub R_positions()
+{
+    my( $self ) = @_;
+    my @vertices = $self->vertices;
+    return grep { !blessed $vertices[$_] && is_chiral_tetrahedral $vertices[$_] &&
+                  $vertices[$_]->{chirality} eq '@@' } 0..$#vertices;
+}
+
+sub S_positions()
+{
+    my( $self ) = @_;
+    my @vertices = $self->vertices;
+    return grep { !blessed $vertices[$_] && is_chiral_tetrahedral $vertices[$_] &&
+                  $vertices[$_]->{chirality} eq '@' } 0..$#vertices;
+}
+
 sub is_main() { $_[0]->{is_main} }
 
 sub is_hydrocarbon()
@@ -239,7 +270,7 @@ sub max_valence()
 
     my $max_valence = 0;
     for my $vertex ($self->vertices) {
-        my $element = ChemOnomatopist::element( $vertex );
+        my $element = ChemOnomatopist::Util::element( $vertex );
         next unless $element;
         next if !exists $elements{$element};
         next if !exists $elements{$element}->{standard_bonding_number};
@@ -358,7 +389,8 @@ sub needs_heteroatom_names()
     }
 
     # Chalcogen analogues of ethers
-    if( @vertices == 1 && $self->parent && grep { ChemOnomatopist::element( @vertices ) eq $_ } qw( S Se Te ) ) {
+    if( @vertices == 1 && $self->parent &&
+        grep { ChemOnomatopist::Util::element( @vertices ) eq $_ } qw( S Se Te ) ) {
         return '';
     }
 
@@ -381,8 +413,8 @@ sub needs_suffix_locant()
     return '' unless @most_senior_groups;
     return 1 if $self->number_of_heteroatoms; # P-15.4.3.2.3: Characteristic groups cited as suffixes are given locants
     return 1 if @most_senior_groups > 2;
-    return 1 if !ChemOnomatopist::element( $most_senior_groups[0] );
-    return 1 if  ChemOnomatopist::element( $most_senior_groups[0] ) ne 'C';
+    return 1 if !ChemOnomatopist::Util::element( $most_senior_groups[0] );
+    return 1 if  ChemOnomatopist::Util::element( $most_senior_groups[0] ) ne 'C';
     return '';
 }
 
@@ -432,7 +464,7 @@ sub heteroatoms()
 {
     my( $self ) = @_;
     my @vertices = $self->vertices;
-    return map { ChemOnomatopist::element( $vertices[$_] ) }
+    return map { ChemOnomatopist::Util::element( $vertices[$_] ) }
                $self->heteroatom_positions;
 }
 
@@ -455,7 +487,7 @@ sub isotopes()
         next if blessed $vertices[$i];
 
         if( exists $vertices[$i]->{isotope} ) {
-            push @isotopes, ChemOnomatopist::Isotope->new( ChemOnomatopist::element( $vertices[$i] ),
+            push @isotopes, ChemOnomatopist::Isotope->new( ChemOnomatopist::Util::element( $vertices[$i] ),
                                                            $vertices[$i]->{isotope},
                                                            $i,
                                                            $self->locants( $i ) );
@@ -683,6 +715,85 @@ sub isotope_part()
     return ChemOnomatopist::Name::Part::Isotope->new( "($isotopes)" );
 }
 
+sub stereodescriptor_part()
+{
+    my( $self ) = @_;
+
+    my @vertices = $self->vertices;
+    my $graph = $self->graph;
+    my $double_bond_positions = set( $self->double_bond_positions );
+    my @stereodescriptors;
+    for my $i (0..$#vertices) {
+        next if blessed $vertices[$i];
+
+        if( is_chiral_tetrahedral( $vertices[$i] ) ) {
+            my @chirality_neighbours = @{$vertices[$i]->{chirality_neighbours}};
+            die "cannot process complicated chiral centers\n" unless @chirality_neighbours == 4;
+
+            my $stereodescriptor = { symbol => $vertices[$i]->{chirality} eq '@' ? 'S' : 'R' };
+            $stereodescriptor->{locant} = join ',', $self->locants( $i ) if $self->length > 1;
+            push @stereodescriptors, $stereodescriptor;
+        } else {
+            next unless $double_bond_positions->has( $i );
+            next if $i == $#vertices;
+            next if any { blessed $_ } map { $vertices[$_] } ( $i, $i+1 );
+            my( $atom2, $atom3 ) = sort { $a->{number} <=> $b->{number} }
+                                   map  { $vertices[$_] }
+                                        ( $i, $i+1 );
+            my $atom1 = first { !blessed $_ && is_cis_trans_bond( $graph, $atom2, $_ ) }
+                              $graph->neighbours( $atom2 );
+            my $atom4 = first { !blessed $_ && is_cis_trans_bond( $graph, $atom3, $_ ) }
+                              $graph->neighbours( $atom3 );
+
+            my $stereodescriptor = {};
+            if( $atom1 && $atom4 ) {
+                my $cistrans = 1; # 1 = cis, -1 = trans
+                if( $graph->get_edge_attribute( $atom1, $atom2, 'bond' ) eq
+                    $graph->get_edge_attribute( $atom3, $atom4, 'bond' ) ) {
+                    $cistrans *= -1;
+                }
+                $cistrans *= -1 if $atom1->{number} > $atom2->{number};
+                $cistrans *= -1 if $atom3->{number} > $atom4->{number};
+
+                my $atom1_new = first { 1 }
+                                sort  { ChemOnomatopist::order_by_neighbours( $graph, $atom2, $a, $b ) }
+                                grep  { $_ != $atom3 }
+                                      $graph->neighbours( $atom2 );
+                my $atom4_new = first { 1 }
+                                sort  { ChemOnomatopist::order_by_neighbours( $graph, $atom3, $a, $b ) }
+                                grep  { $_ != $atom2 }
+                                      $graph->neighbours( $atom3 );
+                $cistrans *= -1 if ($atom1 == $atom1_new) ^ ($atom4 == $atom4_new);
+                $stereodescriptor->{symbol} = $cistrans == 1 ? 'Z' : 'E';
+            } else {
+                next; # TODO: Finish the implementation
+                next if $self->isa( ChemOnomatopist::Chain::Monocycle:: ) && $self->length <= 7;
+                next if $self->isa( ChemOnomatopist::Chain::Bicycle:: );
+
+                my @neighbours2 = grep { $_ != $atom3 } $graph->neighbours( $atom2 );
+                my @neighbours3 = grep { $_ != $atom2 } $graph->neighbours( $atom3 );
+                next if all { !blessed $_ && $_->{symbol} eq 'H' } @neighbours2;
+                next if all { !blessed $_ && $_->{symbol} eq 'H' } @neighbours3;
+                $stereodescriptor->{symbol} = 'Ξ';
+            }
+
+            $stereodescriptor->{locant} = join ',', $self->locants( $i ) if $self->length > 2;
+            push @stereodescriptors, $stereodescriptor;
+        }
+    }
+    return ChemOnomatopist::Name->new unless @stereodescriptors;
+
+    if( all { exists $_->{locant} } @stereodescriptors ) {
+        @stereodescriptors = map  { $_->{locant} . $_->{symbol} }
+                             sort { ChemOnomatopist::cmp_locants( $a->{locant}, $b->{locant} ) }
+                                  @stereodescriptors;
+    } else {
+        @stereodescriptors = map { $_->{symbol} } @stereodescriptors;
+    }
+
+    return ChemOnomatopist::Name::Part::Stereodescriptor->new( '(' . join( ',', @stereodescriptors ) . ')-' )->to_name;
+}
+
 sub prefix()
 {
     my( $self ) = @_;
@@ -692,7 +803,7 @@ sub prefix()
         return $vertex->prefix if blessed $vertex;
 
         # Chalcogen analogues of ethers
-        my $element = ChemOnomatopist::element( $vertex );
+        my $element = ChemOnomatopist::Util::element( $vertex );
         return ChemOnomatopist::Name->new( 'sulfan' ) if $element eq 'S';
         return ChemOnomatopist::Name->new( 'selan'  ) if $element eq 'Se';
         return ChemOnomatopist::Name->new( 'tellan' ) if $element eq 'Te';
@@ -711,7 +822,8 @@ sub suffix()
     my @chain = $self->vertices;
 
     my $name = ChemOnomatopist::Name->new;
-    if( $self->length == 1 && !blessed $chain[0] && ChemOnomatopist::element( @chain ) ne 'C' ) {
+    if( $self->length == 1 && !blessed $chain[0] &&
+        ChemOnomatopist::Util::element( @chain ) ne 'C' ) {
         return $name . 'ne'; # Leaving element prefix appending to the caller
     }
 
@@ -723,8 +835,9 @@ sub suffix()
     my @triple = grep { $bonds[$_] eq '#' } 0..$#bonds;
 
     # BBv2 P-63.2.2.2
-    if( $self->parent && @chain && (all { !blessed $_ } @chain) && ChemOnomatopist::element( @chain ) eq 'O' &&
-        !@double && !@triple && all { ChemOnomatopist::element( $_ ) eq 'C' } @chain[1..$#chain] ) {
+    if( $self->parent && @chain && (all { !blessed $_ } @chain) &&
+        ChemOnomatopist::Util::element( @chain ) eq 'O' &&
+        !@double && !@triple && all { ChemOnomatopist::Util::element( $_ ) eq 'C' } @chain[1..$#chain] ) {
         $name->append_stem( ChemOnomatopist::alkane_chain_name( $self->length - 1 ) );
         $name .= 'oxy';
         return $name;
@@ -734,7 +847,7 @@ sub suffix()
         $self->isa( ChemOnomatopist::Chain::Amine:: ) ) {
         $name->append_stem( ChemOnomatopist::alkane_chain_name( scalar grep { !blessed $_ } $self->vertices ) );
     } elsif( (any { ChemOnomatopist::is_element( $_, 'C' ) } @chain) ||
-        scalar( uniq map { ChemOnomatopist::element( $_ ) } @chain ) > 1 ) {
+        scalar( uniq map { ChemOnomatopist::Util::element( $_ ) } @chain ) > 1 ) {
         $name->append_stem( ChemOnomatopist::alkane_chain_name( $self->length ) );
     }
 
