@@ -1,11 +1,11 @@
 package ChemOnomatopist::Chain::Amide;
 
-use strict;
-use warnings;
-
 # ABSTRACT: Amide chain
 # VERSION
 
+use strict;
+use warnings;
+
 use ChemOnomatopist;
 use ChemOnomatopist::Name;
 use List::Util qw( first );
@@ -39,12 +39,6 @@ sub locants(@)
     return map { $_ ? $_ : 'N' } @_;
 }
 
-sub needs_substituent_locants() { return 1 }
-
-sub suffix()
-{
-    my( $self ) = @_;
-    return $self->{chain}->suffix;
-}
+sub needs_substituent_locants() { 1 }
 
 1;
