@@ -9,9 +9,8 @@ use warnings;
 use parent ChemOnomatopist::Chain::Circular::;
 
 use ChemOnomatopist::Name;
-use ChemOnomatopist::Util::Graph qw(
-    graph_without_edge_attributes
-);
+use ChemOnomatopist::Util;
+use ChemOnomatopist::Util::Graph qw( graph_without_edge_attributes );
 use Graph::Nauty qw( are_isomorphic );
 use Graph::Undirected;
 
@@ -35,7 +34,7 @@ sub has_form($$)
 
     return are_isomorphic( graph_without_edge_attributes( $graph ),
                            $class->ideal_graph,
-                           sub { ChemOnomatopist::element( $_[0] ) } );
+                           sub { ChemOnomatopist::Util::element( $_[0] ) } );
 }
 
 sub ideal_graph($)
