@@ -11,6 +11,7 @@ use ChemOnomatopist::Chain;
 use ChemOnomatopist::Chain::ABA;
 use ChemOnomatopist::Group::AcidHalide;
 use ChemOnomatopist::Group::Amidine;
+use ChemOnomatopist::Group::Azide;
 use ChemOnomatopist::Group::Carbaldehyde;
 use ChemOnomatopist::Group::Carbonitrile;
 use ChemOnomatopist::Chain::Carboxamide;
@@ -43,11 +44,12 @@ use ChemOnomatopist::Group::SulfonicAcid;
 use ChemOnomatopist::Group::Sulfonyl;
 use ChemOnomatopist::Group::Urea;
 use ChemOnomatopist::Group::XO3;
+use ChemOnomatopist::Util qw( element );
 use Chemistry::OpenSMILES qw(
     is_double_bond
+    is_triple_bond
 );
 use Graph::Grammar;
-use Graph::MoreUtils qw( graph_replace );
 use List::Util qw( all any first sum );
 use Scalar::Util qw( blessed );
 
@@ -59,22 +61,22 @@ our @EXPORT_OK = qw(
 sub is_nongroup      { !$_[0]->groups( $_[1] ) }
 sub is_nongroup_atom { !blessed $_[1] && !$_[0]->groups( $_[1] ) && exists $_[1]->{symbol} }
 
-sub is_C  { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) eq 'C' }
-sub is_N  { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) eq 'N' }
-sub is_O  { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) eq 'O' }
-sub is_S  { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) eq 'S' }
-sub is_Se { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) eq 'Se' }
-sub is_Te { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) eq 'Te' }
+sub is_C  { element( $_[1] ) && element( $_[1] ) eq 'C' }
+sub is_N  { element( $_[1] ) && element( $_[1] ) eq 'N' }
+sub is_O  { element( $_[1] ) && element( $_[1] ) eq 'O' }
+sub is_S  { element( $_[1] ) && element( $_[1] ) eq 'S' }
+sub is_Se { element( $_[1] ) && element( $_[1] ) eq 'Se' }
+sub is_Te { element( $_[1] ) && element( $_[1] ) eq 'Te' }
 
-sub is_As_N_B_P_Se_Si_Sb_S_Te { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) =~ /^(As|Br|Cl|F|I|Sb|Se|Si|Te|N|B|P|S)$/ }
-sub is_Br_Cl_F_I              { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) =~ /^(Br|Cl|F|I)$/ }
-sub is_Br_Cl_F_I_N            { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) =~ /^(Br|Cl|F|I|N)$/ }
-sub is_B_Cl_F_I               { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) =~ /^(B|Cl|F|I)$/ }
-sub is_S_Se_Te                { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) =~ /^(S|Se|Te)$/ }
-sub is_O_S_Se_Te              { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) =~ /^(O|S|Se|Te)$/ }
-sub is_C_N_O_S_Se_Te          { ChemOnomatopist::element( $_[1] ) && ChemOnomatopist::element( $_[1] ) =~ /^(C|N|O|S|Se|Te)$/ }
+sub is_As_N_B_P_Se_Si_Sb_S_Te { element( $_[1] ) && element( $_[1] ) =~ /^(As|Br|Cl|F|I|Sb|Se|Si|Te|N|B|P|S)$/ }
+sub is_Br_Cl_F_I              { element( $_[1] ) && element( $_[1] ) =~ /^(Br|Cl|F|I)$/ }
+sub is_Br_Cl_F_I_N            { element( $_[1] ) && element( $_[1] ) =~ /^(Br|Cl|F|I|N)$/ }
+sub is_B_Cl_F_I               { element( $_[1] ) && element( $_[1] ) =~ /^(B|Cl|F|I)$/ }
+sub is_S_Se_Te                { element( $_[1] ) && element( $_[1] ) =~ /^(S|Se|Te)$/ }
+sub is_O_S_Se_Te              { element( $_[1] ) && element( $_[1] ) =~ /^(O|S|Se|Te)$/ }
+sub is_C_N_O_S_Se_Te          { element( $_[1] ) && element( $_[1] ) =~ /^(C|N|O|S|Se|Te)$/ }
 
-sub is_heteroatom { ChemOnomatopist::element( $_[1] ) && !&is_C }
+sub is_heteroatom { element( $_[1] ) && !&is_C }
 
 sub charge_plus_one  {  ChemOnomatopist::charge( $_[1] ) ==  1 }
 sub charge_minus_one {  ChemOnomatopist::charge( $_[1] ) == -1 }
@@ -122,20 +124,20 @@ sub looks_like_ABA_chain
         # ABA chains on both sides
         return '' unless $neighbours[0]->outer_element eq $neighbours[1]->outer_element;
         return '' unless $neighbours[0]->inner_element eq $neighbours[1]->inner_element;
-        return '' unless $neighbours[0]->inner_element eq ChemOnomatopist::element( $center );
+        return '' unless $neighbours[0]->inner_element eq element( $center );
     } elsif( any { blessed $_ && $_->isa( ChemOnomatopist::Chain::ABA:: ) } @neighbours ) {
         # ABA chain on one side
         @neighbours = reverse @neighbours if blessed $neighbours[1] &&
                                              $neighbours[1]->isa( ChemOnomatopist::Chain::ABA:: );
         return '' if any { blessed $_ } ( $center, $neighbours[1] );
-        return '' if $neighbours[0]->inner_element eq ChemOnomatopist::element( $center );
-        return '' if $neighbours[0]->outer_element eq ChemOnomatopist::element( $neighbours[1] );
+        return '' if $neighbours[0]->inner_element eq element( $center );
+        return '' if $neighbours[0]->outer_element eq element( $neighbours[1] );
     } else {
         # No ABA chain yet
         return '' if any { blessed $_ } @neighbours;
-        return '' unless ChemOnomatopist::element( $neighbours[0] ) eq ChemOnomatopist::element( $neighbours[1] );
-        my $outer = ChemOnomatopist::element( $neighbours[0] );
-        my $inner = ChemOnomatopist::element( $center );
+        return '' unless element( $neighbours[0] ) eq element( $neighbours[1] );
+        my $outer = element( $neighbours[0] );
+        my $inner = element( $center );
         return '' unless $elements{$outer}->{seniority} > $elements{$inner}->{seniority};
     }
 
@@ -173,11 +175,18 @@ my @rules = (
       \&anything,
       NO_MORE_VERTICES,
 
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Carboxyl->new( $_[2], $_[3] ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Carboxyl->new( $_[2], $_[3] ), @_[1..3] ) } ],
 
     # Nitramide
     [ sub { &is_nongroup_atom && &is_N && &charge_plus_one }, ( sub { &is_nongroup_atom && &is_O && &has_1_neighbour } ) x 2, sub { &is_nongroup_atom && &is_N }, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Nitramide->new, @_[1..4] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Nitramide->new, @_[1..4] ) } ],
+
+    # Azide
+    [ sub { &is_nongroup_atom && &is_N && &charge_plus_one },
+        EDGE { is_double_bond( @_ ) }, sub { &is_nongroup_atom && &is_N && &charge_minus_one && &has_1_neighbour },
+        EDGE { is_double_bond( @_ ) }, \&is_N,
+      NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Azide->new, @_[1..3] ) } ],
 
     # Hydrazine and diazene
     [ ( sub { &is_nongroup_atom && &is_N } ) x 2,
@@ -201,27 +210,27 @@ my @rules = (
     # Amide
     # CHECKME: Why ketone has to be deleted and not used in replace? It fails somewhy.
     [ sub { &is_nongroup_atom && &is_C }, \&is_amine, \&is_ketone,
-      sub { $_[0]->delete_vertices( $_[3] ); graph_replace( $_[0], ChemOnomatopist::Group::Amide->new( $_[1], $_[3] ), $_[2] ) } ],
+      sub { $_[0]->delete_vertices( $_[3] ); $_[0]->replace( ChemOnomatopist::Group::Amide->new( $_[1], $_[3] ), $_[2] ) } ],
     [ sub { &is_sulfinyl || &is_sulfonyl }, \&is_amine,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Amide->new( $_[1] ), $_[2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Amide->new( $_[1] ), $_[2] ) } ],
 
     # Aldehyde
     [ sub { &is_nongroup_atom && &is_C && &has_H1 }, \&is_ketone,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Aldehyde->new( $_[2] ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Aldehyde->new( $_[2] ), @_[1..2] ) } ],
 
     # Aldehydes attached to carbon in cyclic system or a heteroatom
     [ \&is_aldehyde, sub { &is_circular || ( &is_nongroup_atom && &is_heteroatom ) }, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Carbaldehyde->new( $_[1] ), $_[1] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Carbaldehyde->new( $_[1] ), $_[1] ) } ],
 
     # Acid halide
     [ sub { &is_sulfinyl || &is_sulfonyl }, \&is_Br_Cl_F_I,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::AcidHalide->new( $_[1], ChemOnomatopist::element( $_[2] ) ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::AcidHalide->new( $_[1], element( $_[2] ) ), @_[1..2] ) } ],
 
     # Acyl halide
     [ sub { &is_nongroup_atom && &is_C }, sub { &is_nongroup_atom && &is_Br_Cl_F_I }, sub { &is_ketone && &is_O }, \&is_C, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::AcylHalide->new( $_[2] ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::AcylHalide->new( $_[2] ), @_[1..3] ) } ],
     [ sub { &is_nongroup_atom && &is_C }, sub { &is_cyanide || &is_isocyanide || &is_isocyanate }, sub { &is_ketone && &is_O }, \&is_C, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::AcylHalide->new( $_[2] ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::AcylHalide->new( $_[2] ), @_[1..3] ) } ],
 
     # a(ba)n chain
     [ sub { &is_nongroup_atom && &is_heteroatom && &looks_like_ABA_chain }, ( sub { &is_nongroup_atom && &is_heteroatom } ) x 2, NO_MORE_VERTICES,
@@ -238,10 +247,12 @@ my @rules = (
           } ],
 
     # O-based groups
-    [ sub { &is_nongroup_atom && &is_O && ( &has_H1 || &charge_minus_one ) }, \&anything, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Hydroxy->new( $_[1] ), $_[1] ) } ],
-    [ sub { &is_nongroup_atom && &is_O && all { is_double_bond( @_, $_ ) } $_[0]->neighbours( $_[1] ) }, \&anything, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Ketone->new( ChemOnomatopist::element( $_[1] ) ), $_[1] ) } ],
+    [ 'hydroxy', sub { &is_nongroup_atom && &is_O && ( &has_H1 || &charge_minus_one ) }, \&anything, NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Hydroxy->new( $_[1] ), $_[1] ) } ],
+    [ sub { &is_nongroup_atom && &is_O },
+        EDGE { is_double_bond( @_ ) }, \&anything,
+      NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Ketone->new( element( $_[1] ) ), $_[1] ) } ],
 
     # Ester
     [ sub { &is_nongroup_atom && &is_C }, sub { &is_ketone && &is_O }, sub { &is_nongroup_atom && &is_O && &no_charge }, \&is_C, NO_MORE_VERTICES,
@@ -250,14 +261,16 @@ my @rules = (
 
     # Ether
     [ sub { &is_nongroup_atom && &is_O }, ( \&is_C ) x 2, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Ether->new, $_[1] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Ether->new, $_[1] ) } ],
 
     # Hydroxy groups and their chalcogen analogues
     [ sub { &is_nongroup_atom && &is_O_S_Se_Te && ( &has_H1 || &charge_minus_one ) }, \&is_C_N_O_S_Se_Te, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Hydroxy->new( $_[1] ), $_[1] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Hydroxy->new( $_[1] ), $_[1] ) } ],
     # Ketones and their chalcogen analogues
-    [ sub { &is_nongroup_atom && &is_O_S_Se_Te && all { is_double_bond( @_, $_ ) } $_[0]->neighbours( $_[1] ) }, \&anything, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Ketone->new( ChemOnomatopist::element( $_[1] ) ), $_[1] ) } ],
+    [ sub { &is_nongroup_atom && &is_O_S_Se_Te && all { is_double_bond( @_, $_ ) } $_[0]->neighbours( $_[1] ) },
+        EDGE { is_double_bond( @_ ) }, \&anything,
+      NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Ketone->new( element( $_[1] ) ), $_[1] ) } ],
 
     # Urea
     [ sub { &is_nongroup_atom && &is_C }, \&is_ketone, ( sub { &is_nongroup_atom && &is_N } ) x 2, NO_MORE_VERTICES,
@@ -265,32 +278,47 @@ my @rules = (
 
     # Isocyanide
     [ sub { &is_nongroup_atom && &is_C && &has_H0 && &charge_minus_one }, sub { &is_nongroup_atom && &is_N && &has_H0 && &charge_plus_one }, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Isocyanide->new, @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Isocyanide->new, @_[1..2] ) } ],
 
     # Isocyanate
     [ sub { &is_nongroup_atom && &is_C && &has_H0 && &no_charge }, \&is_ketone, \&is_N, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Isocyanate->new( $_[2]->element ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Isocyanate->new( $_[2]->element ), @_[1..3] ) } ],
 
     # N-based groups
-    [ sub { &is_nongroup_atom && &is_N && &has_H0 && &no_charge }, sub { &is_nongroup_atom && &is_C && &no_charge }, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Cyanide->new, @_[1..2] ) } ],
+    [ 'cyanide without N-substitutions',
+      sub { &is_nongroup_atom && &is_N },
+        EDGE { is_triple_bond( @_ ) }, sub { &is_nongroup_atom && &is_C },
+        NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Cyanide->new, @_[1..2] ) } ],
+    [ 'cyanide with N-substitutions',
+      sub { &is_nongroup_atom && &is_N },
+        EDGE { is_triple_bond( @_ ) }, sub { &is_nongroup_atom && &is_C },
+      sub { die "cannot name N-substituted cyano groups yet\n" } ],
+    # TODO: Recognise aminium cations
+    # [ sub { &is_nongroup_atom && &is_N && &has_H0 && &charge_plus_one }, ( \&anything ) x 4, NO_MORE_VERTICES,
+    #   sub { $_[0]->replace( ChemOnomatopist::Group::Amine->new( $_[1] ), $_[1] ) } ],
     [ sub { &is_nongroup_atom && &is_N && &no_charge }, ( \&anything ) x 3, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Amine->new, $_[1] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Amine->new( $_[1] ), $_[1] ) } ],
     [ sub { &is_nongroup_atom && &is_N && &has_H1 }, ( \&anything ) x 2, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Amine->new, $_[1] ) } ],
-    [ sub { &is_nongroup_atom && &is_N && &has_H2 }, \&anything, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Amine->new, $_[1] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Amine->new( $_[1] ), $_[1] ) } ],
+    [ sub { &is_nongroup_atom && &is_N && ( &has_H2 || ( &has_H1 && &charge_minus_one ) ) }, \&anything, NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Amine->new( $_[1] ), $_[1] ) } ],
     [ sub { &is_nongroup_atom && &is_N && &has_H3 }, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Amine->new, $_[1] ) } ],
-    [ sub { &is_nongroup_atom && &is_N && &has_H0 && any { ChemOnomatopist::element( $_ ) eq 'C' && is_double_bond( @_, $_ ) && $_[0]->degree( $_ ) + ($_->{hcount} ? $_->{hcount} : 0) == 3 } $_[0]->neighbours( $_[1] ) }, \&is_C, \&anything, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Imine->new, $_[1] ) } ],
-    [ sub { &is_nongroup_atom && &is_N && &has_H1 && any { ChemOnomatopist::element( $_ ) eq 'C' && is_double_bond( @_, $_ ) && $_[0]->degree( $_ ) + ($_->{hcount} ? $_->{hcount} : 0) == 3 } $_[0]->neighbours( $_[1] ) }, \&is_C, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Imine->new, $_[1] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Amine->new( $_[1] ), $_[1] ) } ],
+    [ sub { &is_nongroup_atom && &is_N && &has_H0 },
+        EDGE { is_double_bond( @_ ) && $_[0]->degree( $_[2] ) + ($_[2]->{hcount} ? $_[2]->{hcount} : 0) == 3 }, \&is_C,
+        \&anything,
+      NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Imine->new( $_[1] ), $_[1] ) } ],
+    [ sub { &is_nongroup_atom && &is_N && ( &has_H1 || ( &has_H0 && &charge_minus_one ) ) },
+        EDGE { is_double_bond( @_ ) && $_[0]->degree( $_[2] ) + ($_[2]->{hcount} ? $_[2]->{hcount} : 0) == 3 }, \&is_C,
+      NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Imine->new( $_[1] ), $_[1] ) } ],
     [ sub { &is_nongroup_atom && &is_N && &charge_plus_one }, \&is_ketone, sub { &is_hydroxy && &is_O && &charge_minus_one }, \&anything,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Nitro->new, @_[1..3] ) } ],
-    [ sub { &is_nongroup_atom && &is_N && &has_H0 && any { ChemOnomatopist::element( $_ ) =~ /^(S|Se|Te)$/ && is_double_bond( @_, $_ ) && $_[0]->degree( $_ ) + ($_->{hcount} ? $_->{hcount} : 0) == 3 } $_[0]->neighbours( $_[1] ) },
-      sub { &is_nongroup_atom && &is_S_Se_Te },
-      \&anything,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Nitro->new, @_[1..3] ) } ],
+    [ sub { &is_nongroup_atom && &is_N && &has_H0 },
+        EDGE { is_double_bond( @_) && $_[0]->degree( $_[2] ) + ($_[2]->{hcount} ? $_[2]->{hcount} : 0) == 3 }, sub { &is_nongroup_atom && &is_S_Se_Te },
+        \&anything,
       NO_MORE_VERTICES,
       sub { $_[0]->add_group( ChemOnomatopist::Chain::Sulfimide->new( @_[0..2] ) ) } ],
 
@@ -301,10 +329,11 @@ my @rules = (
       sub { die "cannot handle nitric/nitrous acid esters yet\n" } ],
 
     # Carbonitrile, special case of cyanide
-    [ \&is_cyanide, \&is_circular, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Carbonitrile->new, $_[1] ) } ],
-    [ \&is_cyanide, \&is_heteroatom, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Carbonitrile->new, $_[1] ) } ],
+    [ 'cyanide attached to circular compound',
+      \&is_cyanide, \&is_circular, NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Carbonitrile->new, $_[1] ) } ],
+    [ 'cyanide attached to heteroatom', \&is_cyanide, \&is_heteroatom, NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Carbonitrile->new, $_[1] ) } ],
 
     # Amidines (BBv3 P-66.4.1)
     [ sub { &is_nongroup_atom && &is_C }, sub { &is_amine && &is_nongroup }, sub { &is_imine && &is_nongroup },
@@ -315,66 +344,66 @@ my @rules = (
       sub { $_[0]->add_group( ChemOnomatopist::Group::Amidine->new( @_[0..4] ) ) } ],
 
     # Nitroso and its analogues
-    [ sub { &is_nongroup_atom && &is_Br_Cl_F_I_N }, \&is_ketone, \&is_C, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Nitroso->new( ChemOnomatopist::element( $_[1] ) ), @_[1..2] ) } ],
+    [ sub { &is_nongroup_atom && &is_Br_Cl_F_I_N }, \&is_ketone, sub { &is_C || &is_N }, NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Nitroso->new( element( $_[1] ) ), @_[1..2] ) } ],
 
     # XO3
     [ sub { &is_nongroup_atom && &is_Br_Cl_F_I }, ( sub { &is_ketone && &is_O } ) x 3,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::XO3->new( ChemOnomatopist::element( $_[1] ) ), @_[1..4] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::XO3->new( element( $_[1] ) ), @_[1..4] ) } ],
 
     # Peroxide
-    [ sub { &is_nongroup_atom && &is_O }, sub { ( &is_nongroup_atom && &is_O ) || ( &is_hydroxy && &charge_minus_one ) }, sub { &is_C }, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Peroxide->new( @_[1..2] ), @_[1..2] ) } ],
+    [ sub { &is_nongroup_atom && &is_O }, sub { ( &is_nongroup_atom && &is_O ) || ( &is_hydroxy && &charge_minus_one ) }, \&is_C, NO_MORE_VERTICES,
+      sub { $_[0]->replace( ChemOnomatopist::Group::Peroxide->new( @_ ), @_[1..2] ) } ],
 
     # Hydroperoxide
     [ sub { &is_nongroup_atom && &is_O_S_Se_Te }, \&is_hydroxy, \&anything, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Hydroperoxide->new( $_[1], $_[2] ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Hydroperoxide->new( $_[1], $_[2] ), @_[1..2] ) } ],
     [ sub { &is_nongroup_atom && &is_O_S_Se_Te }, sub { &is_nongroup_atom && &is_O && &charge_minus_one }, \&anything, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Hydroperoxide->new( $_[1], $_[2] ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Hydroperoxide->new( $_[1], $_[2] ), @_[1..2] ) } ],
 
     # S-based groups
     [ sub { &is_nongroup_atom && &is_S_Se_Te }, sub { &is_ketone || ( &is_nongroup_atom && &is_N && &has_H1 ) || &is_hydrazine }, sub { &is_hydroxy || &is_hydroperoxide }, \&is_C, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::SulfinicAcid->new( ChemOnomatopist::element( $_[1] ), @_[2..3] ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::SulfinicAcid->new( element( $_[1] ), @_[2..3] ), @_[1..3] ) } ],
     [ sub { &is_nongroup_atom && &is_S_Se_Te }, ( sub { &is_ketone || ( &is_nongroup_atom && &is_N && &has_H1 ) || &is_hydrazine } ) x 2, sub { &is_hydroxy || &is_hydroperoxide }, \&is_C, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::SulfonicAcid->new( ChemOnomatopist::element( $_[1] ), @_[2..4] ), @_[1..4] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::SulfonicAcid->new( element( $_[1] ), @_[2..4] ), @_[1..4] ) } ],
 
     # Sulfoxide group and its analogues
     [ sub { &is_nongroup_atom && &is_S_Se_Te }, \&is_ketone, ( \&anything ) x 2, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Sulfinyl->new( ChemOnomatopist::element( $_[1] ) ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Sulfinyl->new( element( $_[1] ), $_[2] ), @_[1..2] ) } ],
     [ sub { &is_nongroup_atom && &is_S_Se_Te }, ( \&is_ketone ) x 2, ( \&anything ) x 2, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Sulfonyl->new( ChemOnomatopist::element( $_[1] ) ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Sulfonyl->new( element( $_[1] ), $_[2], $_[3] ), @_[1..3] ) } ],
 
     # Noncarbon oxoacids
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, ( sub { &is_hydroxy && &is_O } ) x 4,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..5] ), @_[1..5] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..5] ), @_[1..5] ) } ],
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, ( sub { &is_hydroxy && &is_O } ) x 2, ( sub { &is_ketone && &is_O } ) x 2,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..5] ), @_[1..5] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..5] ), @_[1..5] ) } ],
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, ( sub { &is_hydroxy && &is_O } ) x 3, sub { &is_ketone && &is_O },
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..5] ), @_[1..5] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..5] ), @_[1..5] ) } ],
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, ( sub { &is_hydroxy && &is_O } ) x 3,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..4] ), @_[1..4] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..4] ), @_[1..4] ) } ],
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, ( sub { &is_hydroxy && &is_O } ) x 2, sub { &is_ketone && &is_O },
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..4] ), @_[1..4] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..4] ), @_[1..4] ) } ],
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, ( sub { &is_ketone && &is_O }  ) x 2, sub { &is_hydroxy && &is_O },
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..4] ), @_[1..4] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..4] ), @_[1..4] ) } ],
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, ( sub { &is_hydroxy && &is_O } ) x 2,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..3] ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..3] ), @_[1..3] ) } ],
     [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te },   sub { &is_hydroxy && &is_O }, sub { &is_ketone && &is_O },
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..3] ), @_[1..3] ) } ],
-    [ sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te },   sub { &is_hydroxy && &is_O },
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..2] ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..3] ), @_[1..3] ) } ],
+    [ 'noncarbon oxoacid', sub { &is_nongroup_atom && &is_As_N_B_P_Se_Si_Sb_S_Te }, sub { &is_hydroxy && &is_O },
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..2] ), @_[1..2] ) } ],
     [ \&is_XO3, sub { &is_hydroxy && &is_O },
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..2] ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..2] ), @_[1..2] ) } ],
     [ sub { &is_sulfinyl || &is_sulfonyl }, ( sub { &is_hydroxy && &is_O } ) x 2,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..3] ), @_[1..3] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..3] ), @_[1..3] ) } ],
     [ \&is_nitro, sub { &is_hydroxy && &is_O },
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..2] ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::NoncarbonOxoacid->new( @_[1..2] ), @_[1..2] ) } ],
 
     # Sulfinamides and sulfonamides
     [ \&is_amide, \&is_sulfinyl, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Sulfinamide->new( $_[2]->element ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Sulfinamide->new( $_[2]->element, $_[2]->{ketone} ), @_[1..2] ) } ],
     [ \&is_amide, \&is_sulfonyl, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], ChemOnomatopist::Group::Sulfonamide->new( $_[2]->element ), @_[1..2] ) } ],
+      sub { $_[0]->replace( ChemOnomatopist::Group::Sulfonamide->new( $_[2]->element, @{$_[2]->{ketones}} ), @_[1..2] ) } ],
 
     # Detecting amides attached to cyclic chains
     [ sub { &is_nongroup_atom && &is_C && 1 == grep { blessed $_ && $_->isa( ChemOnomatopist::Group::Amide:: ) && $_->{parent} == $_[1] } $_[0]->neighbours( $_[1] ) }, \&is_amide, \&is_monocycle, NO_MORE_VERTICES,
@@ -387,9 +416,9 @@ my @rules = (
 # Old unused rules
 my @rules_old = (
     [ \&is_C, \&is_benzene, \&is_ketone, \&is_N, NO_MORE_VERTICES,
-      sub { graph_replace( $_[0], { type => 'benzamide' }, @_[1..4] ) } ],
+      sub { $_[0]->replace( { type => 'benzamide' }, @_[1..4] ) } ],
 
-    [ \&is_benzene, \&is_hydroxy, sub { graph_replace( $_[0], { type => 'phenol' }, @_[1..2] ) } ],
+    [ \&is_benzene, \&is_hydroxy, sub { $_[0]->replace( { type => 'phenol' }, @_[1..2] ) } ],
 );
 
 sub is_mainchain() { any { $_->is_main } $_[0]->groups( $_[1] ) }
@@ -448,7 +477,7 @@ our @mainchain_rules = (
     # Unpack sidechain amides
     [ sub { !&is_mainchain && &is_amide && !&is_carboxamide },
       sub {
-            graph_replace( $_[0], ChemOnomatopist::Group::Amine->new, $_[1] );
+            $_[0]->replace( ChemOnomatopist::Group::Amine->new, $_[1] );
             $_[0]->set_edge_attribute( $_[1]->{parent}, $_[1]->{ketone}, 'bond', '=' ) if $_[1]->{ketone};
           } ],
 
